/*
 * Load_symmod.cpp
 * ---------------
 * Purpose: SymMOD (Symphonie Pro) module loader
 * Notes  : Some effects aren't supported.
 *          Virtual instruments are partially supported.
 *          DSP settings aren't supported (yet?), but instruments which use the DSP will be set to use
 *          plugin number 1, so you can easily configure a suitable effect yourself in OpenMPT.
 *          Based in part on Patrick Meng's Java-based Symphonie player and its source.
 *          Some effect behavior and other things are based on the original Amiga assembly source.
 * Authors: Devin Acker
 *          OpenMPT Devs
 * The OpenMPT source code is released under the BSD license. Read LICENSE for more details.
*/

#include "stdafx.h"
#include "Loaders.h"
#include "modsmp_ctrl.h"

#include <map>

#ifdef LIBOPENMPT_BUILD
// TODO fix
//#define MPT_SYMMOD_USE_REAL_SUBSONGS
#endif

OPENMPT_NAMESPACE_BEGIN

struct SymEvent
{
	enum Command : uint8
	{
		KeyOn = 0,
		VolSlideUp,
		VolSlideDown,
		PitchSlideUp,
		PitchSlideDown,
		ReplayFrom,
		FromAndPitch,
		SetFromAdd,
		FromAdd,
		SetSpeed,
		AddPitch,
		AddVolume,
		Vibrato,
		Tremolo,
		SampleVib,
		PitchSlideTo,
		Retrig,
		Emphasis,
		AddHalfTone,
		CV,
		CVAdd,

		Filter = 23,
		DSPEcho,
		DSPDelay,
	};
	uint8be command;
	int8be  note;
	enum Volume
	{
		VolCommand = 200,
		StopSample = 254,
		ContSample = 253,
		StartSample = 252, // unused
		KeyOff = 251,
		SpeedDown = 250,
		SpeedUp = 249,
		SetPitch = 248,
		PitchUp = 247,
		PitchDown = 246,
		PitchUp2 = 245,
		PitchDown2 = 244,
		PitchUp3 = 243,
		PitchDown3 = 242
	};
	uint8be param;
	uint8be inst;

	// used to compare DSP events for mapping them to MIDI macro numbers
	bool operator<(const SymEvent &other) const
	{
		return std::tie(command, note, param, inst) < std::tie(other.command, other.note, other.param, other.inst);
	}
};

MPT_BINARY_STRUCT(SymEvent, 4);


// Virtual instrument info
// This allows instruments to be created based on a mix of other instruments.
// The sample mixing is done at load time.
struct SymVirtualInst
{
	char    id[4];  // "ViRT"
	uint8be dummy[10];

	uint16be numEvents;
	uint16be maxEvents; // always 20
	uint16be eventSize; // always 4

	SymEvent events[20];

	char padding[28];
};

MPT_BINARY_STRUCT(SymVirtualInst, 128);


struct SymFilterSetting
{
	uint8be cutoff;
	uint8be resonance;
};

MPT_BINARY_STRUCT(SymFilterSetting, 2);


// Instrument definition
struct SymInstrument
{
	struct SymInstrumentName
	{
		char name[128];
	};
	SymVirtualInst virt;  // or SymInstrumentName

	enum Type : int8
	{
		Silent  = -8,
		Kill    = -4,
		Normal  = 0,
		Loop    = 4,
		Sustain = 8
	};
	int8be   type;
	uint8be  loopStartHigh;
	uint8be  loopLenHigh;
	uint8be  numLoops; // for "sustain" instruments
	enum Channels
	{
		Mono,
		StereoL,
		StereoR,
		LineSrc // virtual mix instrument
	};
	uint8be  channels;
	uint8be  dummy1; // called "automaximize" (normalize?) in amiga source, but unused
	uint8be  volume; // 0-200
	uint8be  dummy2[3]; // info about "parent/child" and sample format
	int8be   finetune; // -128..127 ~= 2 semitones
	int8be   transpose;
	enum SampleFlags
	{
		PlayReverse   = 1, // reverse sample
		AsQueue       = 2, // "queue" virtual instrument (rendered track)
		MirrorX       = 4, // phase invert sample
		Is16Bit       = 8, // not used, we already know the bit depth of the samples
		NewLoopSystem = 16, // use fine loop start/len values

		MakeNewSample = (PlayReverse | MirrorX)
	};
	uint8be  sampleFlags;
	uint8be  filterType; // maybe same as filter events (0 = none, 1 = lowpass, 2 = highpass, 3 = bandpass)?
	enum InstFlags
	{
		NoTranspose = 1, // don't apply sequence/position transpose
		NoDSP       = 2, // don't apply DSP effects
		SyncPlay    = 4  // play a stereo instrument pair on consecutive channels
	};
	uint8be  instFlags;
	uint8be  downsample; // downsample factor; affects sample tuning
	uint8be  dummy4[2];  // resonance, "loadflags" (both unused)
	uint8be  info; // ? (only lowest bit used?)
	uint8be  rangeStart; // ?
	uint8be  rangeLen; // ?
	uint8be  dummy5;
	uint16be loopStartFine;
	uint16be loopLenFine;
	uint8be  dummy6[6];

	uint8be  filterFlags; // ?
	uint8be  filterPoints; // # of filter envelope points (up to 4, possibly only 1-2 ever actually used)
	SymFilterSetting filter[4];

	uint8be  padding[86];

	bool IsVirtual() const
	{
		return !memcmp(virt.id, "ViRT", 4)
			&& virt.eventSize == sizeof(SymEvent)
			&& virt.numEvents <= virt.maxEvents
			&& virt.maxEvents <= std::size(virt.events);
	}

	bool HasName() const
	{
		return virt.id[0] != 0;
	}

	void ConvertToMPT(ModInstrument &mptInst, ModSample &mptSmp, CSoundFile &sndFile) const
	{
		// name
		if(!IsVirtual())
			mptInst.name = mpt::String::ReadBuf(mpt::String::maybeNullTerminated, mpt::bit_cast<SymInstrumentName>(virt).name);

		// loop
		// TODO: figure out why loop points not usually being set correctly
		SmpLength loopStart = static_cast<SmpLength>(loopStartHigh) << 16;
		SmpLength loopLen = (static_cast<SmpLength>(loopLenHigh) << 16);
		if(sampleFlags & NewLoopSystem)
		{
			loopStart += loopStartFine;
			loopLen += loopLenFine;
		}
		const double loopScale = static_cast<double>(mptSmp.nLength) / (100 << 16);
		loopStart = mpt::saturate_cast<SmpLength>(loopStart * loopScale);
		loopLen = std::min(mptSmp.nLength, mpt::saturate_cast<SmpLength>(loopLen * loopScale));
		if(type == Loop)
		{
			mptSmp.uFlags.set(CHN_LOOP);

			mptSmp.nLoopStart = loopStart;
			mptSmp.nLoopEnd = loopStart + loopLen;
		} else if(type == Sustain && numLoops > 1)
		{
			// TODO: set up multiple loop
			mptSmp.uFlags.reset(CHN_LOOP);
		} else
		{
			mptSmp.uFlags.reset(CHN_LOOP);
		}

		// volume (0-200, default 100)
		// TODO: amplify sample if >100? (Symphonie does)
		if (volume > 0 && volume <= 200)
			mptSmp.nGlobalVol = std::min<uint16>(volume * 64 / 200, 64);
		else
			mptSmp.nGlobalVol = 32u;

		// apply reverse/invert flags
		if (sampleFlags & PlayReverse)
		{
			mptSmp.uFlags.set(CHN_REVERSE);
			//ctrlSmp::ReverseSample(mptSmp, 0, 0, module);
		}
		if (sampleFlags & MirrorX)
		{
			ctrlSmp::InvertSample(mptSmp, 0, 0, sndFile);
		}

		// tuning info
		// importing a sample of A440 at 44.1kHz and playing it at the closest possible note in Symphonie yields ~427.64Hz
		// which is almost exactly half a semitone lower. if we assume an ideal base sample rate of 44.1kHz, then we should
		// detune it by the same amount in order to correctly match the original tracker
		// TODO: why do some songs sound badly out of tune with the winamp plugin (or maybe all libopenmpt?)
		mptSmp.nC5Speed = 42862u;
		mptSmp.Transpose(-downsample + (transpose / 12.0) + (finetune / (128.0 * 12.0)));

#ifndef NO_PLUGINS
		// DSP settings
		mptInst.nMixPlug = (instFlags & NoDSP) ? 0 : 1;
#endif

		// filter settings
		if(filterFlags != 0)
		{
			/*
			mptInst.SetCutoff(0x7F, true);
			// just average both resonance values since we can't interpolate them
			mptInst.SetResonance((resonance1 + resonance2) >> 2, true);

			// create filter sweep envelope
			// Symphonie applies this to the actual sample, which means the sweep rate depends
			// on the sample length and the note being played, which we can't really do here
			InstrumentEnvelope &filterEnv = mptInst.GetEnvelope(EnvelopeType::ENV_PITCH);
			filterEnv.dwFlags.set(ENV_ENABLED | ENV_FILTER);

			// TODO: figure out a good way to fake the length
			uint16 sweepLen = 1 + (mptSmp.nLength / 350);
			filterEnv.push_back(EnvelopeNode(0, cutoff1 >> 2));
			filterEnv.push_back(EnvelopeNode(sweepLen, cutoff2 >> 2));
			*/

			mptInst.SetCutoff(filter[0].cutoff >> 1, true);
			mptInst.SetResonance(filter[0].resonance >> 1, true);
		} else if(instFlags & NoDSP)
		{
			mptInst.SetCutoff(127, true);
			mptInst.SetResonance(0, true);
		}
	}
};

MPT_BINARY_STRUCT(SymInstrument, 256);


struct SymSequence
{
	uint16be start;
	uint16be length;
	uint16be loop;
	int16be  info;
	int16be  transpose;

	uint8be  padding[6];
};

MPT_BINARY_STRUCT(SymSequence, 16);

struct SymPosition
{
	uint8be  dummy[4];
	uint16be loopNum;
	uint16be loopCount; // unused
	uint16be pattern;
	uint16be start;
	uint16be length;
	uint16be speed;
	int16be  transpose;
	uint16be eventsPerLine; // not used, hopefully

	uint8be  padding[12];

	// used to compare position entries for mapping them to OpenMPT patterns
	bool operator<(const SymPosition &other) const
	{
		return std::tie(pattern, start, length, transpose, speed, loopNum) < std::tie(other.pattern, other.start, other.length, other.transpose, other.speed, other.loopNum);
	}
};

MPT_BINARY_STRUCT(SymPosition, 32);


static std::vector<std::byte> DecodeSymChunk(FileReader &file)
{
	std::vector<std::byte> data;
	const uint32 packedLength = file.ReadUint32BE();

	FileReader chunk = file.ReadChunk(packedLength);

	if(!chunk.CanRead(packedLength))
		return data;

	if(chunk.ReadMagic("PACK\xFF\xFF"))
	{
		// Compressed chunk
		uint32 unpackedLength = chunk.ReadUint32BE();
		// The best compression ratio can be achieved with type 3, where one bytes turns into up to 255 bytes.
		MPT_ASSERT(unpackedLength <= packedLength * 255);
		uint32 maxLength = packedLength;
		if(Util::MaxValueOfType(maxLength) / 255 >= maxLength)
			maxLength *= 255;
		else
			maxLength = Util::MaxValueOfType(maxLength);
		LimitMax(unpackedLength, maxLength);
		data.resize(unpackedLength);

		bool done = false;
		uint32 offset = 0, remain = unpackedLength;

		while(!done && !chunk.EndOfFile())
		{
			uint8 len;
			std::byte dword[4];

			const int8 type = chunk.ReadInt8();
			switch(type)
			{
			case 0:
				// copy raw bytes
				len = chunk.ReadUint8();
				if(remain < len|| !chunk.CanRead(len))
				{
					done = true;
				} else
				{
					chunk.ReadRaw(mpt::as_span(data).subspan(offset, len));
					offset += len;
					remain -= len;
				}
				break;

			case 1:
				// copy a dword multiple times
				len = chunk.ReadUint8();
				if(remain < (len * 4u) || !chunk.CanRead(4))
				{
					done = true;
				} else
				{
					chunk.ReadArray(dword);
					while(len--)
					{
						data[offset++] = dword[0];
						data[offset++] = dword[1];
						data[offset++] = dword[2];
						data[offset++] = dword[3];
						remain -= 4;
					}
				}
				break;

			case 2:
				// copy a dword twice
				if(remain < 8 || !chunk.CanRead(4))
				{
					done = true;
				} else
				{
					chunk.ReadArray(dword);
					data[offset++] = dword[0];
					data[offset++] = dword[1];
					data[offset++] = dword[2];
					data[offset++] = dword[3];
					data[offset++] = dword[0];
					data[offset++] = dword[1];
					data[offset++] = dword[2];
					data[offset++] = dword[3];
					remain -= 8;
				}
				break;

			case 3:
				// zero bytes
				len = chunk.ReadUint8();
				if(remain < len)
				{
					done = true;
				} else
				{
					// vector is already initialized to zero
					offset += len;
					remain -= len;
				}
				break;

			case -1:
				done = true;
				break;

			default:
				// error
				done = true;
				break;
			}
		}

		if(remain)
			data.clear();
	} else
	{
		// Uncompressed chunk
		if(!chunk.ReadVector(data, packedLength))
			data.clear();
	}

	return data;
}


template<typename T>
static std::vector<T> DecodeSymArray(FileReader &file)
{
	const auto data = DecodeSymChunk(file);
	FileReader chunk(mpt::as_span(data));
	std::vector<T> retVal;
	chunk.ReadVector(retVal, data.size() / sizeof(T));
	return retVal;
}


static bool ReadRawSymSample(ModSample &sample, FileReader &file)
{
	SampleIO::Bitdepth depth = SampleIO::_8bit;
	SampleIO::Channels channels = SampleIO::mono;

	sample.Initialize();

	file.Rewind();
	if(file.ReadMagic("MAESTRO"))
	{
		depth = SampleIO::_16bit;

		file.Seek(12);
		if(file.ReadUint32BE() == 0)
		{
			channels = SampleIO::stereoInterleaved;
			sample.nLength = file.BytesLeft() / 4;
		} else
		{
			sample.nLength = file.BytesLeft() / 2;
		}
	} else if(file.ReadMagic("16BT"))
	{
		depth = SampleIO::_16bit;

		file.Seek(6);
		sample.nLength = file.BytesLeft() / 2;
	} else
	{
		sample.nLength = file.BytesLeft();
	}

	return SampleIO(
		depth, channels,
		SampleIO::bigEndian,
		SampleIO::signedPCM)
		.ReadSample(sample, file) > 0;
}


static std::vector<std::byte> DecodeSample8(FileReader &file)
{
	auto data = DecodeSymChunk(file);
	uint8 lastVal = 0;
	for(auto &val : data)
	{
		lastVal += mpt::byte_cast<uint8>(val);
		val = mpt::byte_cast<std::byte>(lastVal);
	}

	return data;
}


static std::vector<std::byte> DecodeSample16(FileReader &file)
{
	auto data = DecodeSymChunk(file);
	std::byte buf[4096];
	// Size of block in 16-bit samples
	constexpr size_t blockSize = sizeof(buf) / 2;

	for(size_t block = 0; block < data.size() / sizeof(buf); block++)
	{
		const size_t offset = block * sizeof(buf);
		uint8 lastVal = 0;

		// Decode LSBs
		for(int i = 0; i < blockSize; i++)
		{
			lastVal += mpt::byte_cast<uint8>(data[offset + i]);
			buf[i * 2 + 1] = mpt::byte_cast<std::byte>(lastVal);
		}
		// Decode MSBs
		for(int i = 0; i < blockSize; i++)
		{
			lastVal += mpt::byte_cast<uint8>(data[offset + i + blockSize]);
			buf[i * 2] = mpt::byte_cast<std::byte>(lastVal);
		}

		std::memcpy(data.data() + offset, buf, sizeof(buf));
	}

	return data;
}


static void NormalizeSample(ModSample &sample)
{
	// pretty much lifted directly from Ctrl_smp.cpp
	SmpLength end = sample.nLength * sample.GetNumChannels();

	if(sample.uFlags[CHN_16BIT])
	{
		int16 *p = sample.sample16();
		int max = 1;
		for (SmpLength i = 0; i < end; i++)
		{
			if (p[i] > max) max = p[i];
			if (-p[i] > max) max = -p[i];
		}
		if (max < 32767)
		{
			max++;
			for (SmpLength j = 0; j < end; j++)
			{
				int l = (((int)p[j]) << 15) / max;
				p[j] = (int16)l;
			}
		}
	} else
	{
		int8 *p = sample.sample8();
		int max = 1;
		for (SmpLength i = 0; i < end; i++)
		{
			if (p[i] > max) max = p[i];
			if (-p[i] > max) max = -p[i];
		}
		if (max < 127)
		{
			max++;
			for (SmpLength j = 0; j < end; j++)
			{
				int l = (((int)p[j]) << 7) / max;
				p[j] = (int8)l;
			}
		}
	}
}


static void ConvertDSP(const SymEvent &event, MIDIMacroConfigData::Macro &str)
{
	if(event.command == SymEvent::Filter)
	{
		// TODO: this is obviously shit, figure out the real filter params
		// Also check the Java player source to see how they do DSPs

		//
		// CFILTER_MAXRESO	EQU	185
		// CFILTER_MAXFREQ	EQU	240
		/*
		uint8 cutoff = 0x40 + (event.param >> 2);
		uint8 reso = event.inst >> 1;

		uint8 cutoff = std::min<uint8>(127, event.param * 127 / 240);
		uint8 reso = std::min<uint8>(127, event.inst * 127 / 185);
		*/

		// uint8 cutoff = std::min<uint8>(127, (event.param * 36000 / 28867) >> 1);
		uint8 cutoff = std::min<uint8>(127, event.param * 3 / 2);
		uint8 reso = std::min<uint8>(127, event.inst >> 1);

		if(event.note == 1)  // lowpass filter
		{
			// TODO: make sure this is the right way around
			std::sprintf(str, "F0F000%02X F0F001%02X F0F00200", cutoff, reso);
		} else if(event.note == 2)  // highpass filter
		{
			std::sprintf(str, "F0F000%02X F0F001%02X F0F00210", cutoff, reso);
		} else  // no filter or unsupported filter type
		{
			std::strcpy(str, "F0F0007F F0F00100");
		}
	}
	// TODO: others
}

bool CSoundFile::ReadSymMOD(FileReader &file, ModLoadingFlags loadFlags)
{
	file.Rewind();
	if(!file.ReadMagic("SymM")
	   || file.ReadUint32BE() != 1)
		return false;

	if (loadFlags == onlyVerifyHeader)
		return true;

	InitializeGlobals(MOD_TYPE_IT/*MOD_TYPE_SYMMOD*/);

	m_SongFlags.set(SONG_LINEARSLIDES);
	// TODO: find a fix for this, this flag gets set again when converting to IT internally so setting it here is useless
	m_playBehaviour.reset(kITPatternLoopTargetReset);

	// TODO: initialize other defaults here

	bool ok = true;

	// Data chunk types
	enum class ChunkType : int32
	{
		NumChannels     = -1,
		TrackLength     = -2,
		PatternSize     = -3,
		NumInstruments  = -4,
		EventSize       = -5,
		Tempo           = -6,
		ExternalSamples = -7,
		PositionList    = -10,
		SampleFile      = -11,
		EmptySample     = -12,
		PatternEvents   = -13,
		InstrumentList  = -14,
		Sequences       = -15,
		InfoText        = -16,
		SamplePacked    = -17,
		SamplePacked16  = -18,
		InfoType        = -19,
		InfoBinary      = -20,
		InfoString      = -21,

		SampleBoost     = 10,
		StereoDetune    = 11,
		StereoPhase     = 12,
	};

	uint32 trackLen = 0;

	std::vector<SymPosition> positions;
	std::vector<SymSequence> sequences;
	std::vector<SymEvent> patternData;
	std::vector<SymInstrument> instruments;

	// map Symphonie instruments to OpenMPT instruments and samples
	std::vector<SAMPLEINDEX> sampleMap;
	std::vector<INSTRUMENTINDEX> instrumentMap;

	while(ok && !file.EndOfFile())
	{
		const ChunkType chunkType = static_cast<ChunkType>(file.ReadInt32BE());
		switch(chunkType)
		{
		// Simple values
		case ChunkType::NumChannels:
			// TODO check against max
			if(auto numChannels = static_cast<CHANNELINDEX>(file.ReadUint32BE()); !m_nChannels && numChannels > 0)
			{
				m_nChannels = numChannels;
				m_nSamplePreAmp = Clamp(512 / m_nChannels, 16, 128);
			}
			break;

		case ChunkType::TrackLength:
			// TODO check against max
			trackLen = file.ReadUint32BE();
			break;

		case ChunkType::EventSize:
			ok = (file.ReadUint32BE() & 0xFFFF) <= sizeof(SymEvent);
			break;

		case ChunkType::Tempo:
			// not 100% accurate, but...
			m_nDefaultTempo = TEMPO(1.24 * file.ReadUint32BE());
			break;

		// Unused values
		case ChunkType::NumInstruments: // determined from # of instrument headers instead
		case ChunkType::PatternSize:
		case ChunkType::ExternalSamples:
		case ChunkType::SampleBoost:
		case ChunkType::StereoDetune:
		case ChunkType::StereoPhase:
			file.Skip(4);
			break;

		// Binary chunk types
		case ChunkType::PositionList:
			if((loadFlags & loadPatternData) && positions.empty())
			{
				positions = DecodeSymArray<SymPosition>(file);
				ok = !positions.empty();
			} else
			{
				file.Skip(file.ReadUint32BE());
			}
			break;

		case ChunkType::SampleFile:
		case ChunkType::SamplePacked:
		case ChunkType::SamplePacked16:
			if((loadFlags & loadSampleData) && CanAddMoreSamples())
			{
				SAMPLEINDEX sample = GetNextFreeSample();

				if(sample != SAMPLEINDEX_INVALID)
				{
					m_nSamples = std::max(sample, m_nSamples);
					sampleMap.push_back(sample);

					std::vector<std::byte> unpackedSample;
					FileReader chunk;
					if(chunkType == ChunkType::SampleFile)
					{
						chunk = file.ReadChunk(file.ReadUint32BE());
					} else if(chunkType == ChunkType::SamplePacked)
					{
						unpackedSample = DecodeSample8(file);
						chunk = mpt::as_span(unpackedSample);
					} else  // SamplePacked16
					{
						unpackedSample = DecodeSample16(file);
						chunk = mpt::as_span(unpackedSample);
					}

					ok = ReadIFFSample(sample, chunk)
					     || ReadWAVSample(sample, chunk)
					     || ReadAIFFSample(sample, chunk)
					     || ReadRawSymSample(Samples[sample], chunk);

					if(!ok)
						break;

					// Symphonie normalizes samples at load time
					NormalizeSample(Samples[sample]);

					// Symphonie represents stereo instruments as two consecutive mono instruments which are
					// automatically played at the same time. if this one uses a stereo sample, split it
					// and map two OpenMPT instruments to the stereo halves to ensure correct playback
					if(Samples[sample].uFlags[CHN_STEREO] && CanAddMoreSamples())
					{
						SAMPLEINDEX sampleR = GetNextFreeSample();

						if(sampleR != SAMPLEINDEX_INVALID)
						{
							m_nSamples = std::max(sampleR, m_nSamples);
							ReadSampleFromSong(sampleR, *this, sample);
							ctrlSmp::ConvertToMono(Samples[sample], *this, ctrlSmp::onlyLeft);
							ctrlSmp::ConvertToMono(Samples[sampleR], *this, ctrlSmp::onlyRight);

							sampleMap.push_back(sampleR);
						} else
						{
							// just map the existing stereo sample again
							sampleMap.push_back(sample);
						}
					}
				} else
				{
					sampleMap.push_back(0);
				}
			} else
			{
				// skip sample
				sampleMap.push_back(0);
				file.Skip(file.ReadUint32BE());
			}
			break;

		case ChunkType::EmptySample:
			sampleMap.push_back(0);
			break;

		case ChunkType::PatternEvents:
			if((loadFlags & loadPatternData) && patternData.empty())
			{
				patternData = DecodeSymArray<SymEvent>(file);
				ok = !patternData.empty();
			} else
			{
				file.Skip(file.ReadUint32BE());
			}
			break;

		case ChunkType::InstrumentList:
			if((loadFlags & loadSampleData) && instruments.empty())
			{
				instruments = DecodeSymArray<SymInstrument>(file);
				ok = !instruments.empty();
			} else
			{
				file.Skip(file.ReadUint32BE());
			}
			break;

		case ChunkType::Sequences:
			if((loadFlags & loadPatternData) && sequences.empty())
			{
				sequences = DecodeSymArray<SymSequence>(file);
				ok = !sequences.empty();
			} else
			{
				file.Skip(file.ReadUint32BE());
			}
			break;

		case ChunkType::InfoText:
		{
			const auto text = DecodeSymChunk(file);
			ok = m_songMessage.Read(text.data(), text.size(), SongMessage::leLF);
			break;
		}

		// Unused binary chunks
		case ChunkType::InfoType:
		case ChunkType::InfoBinary:
		case ChunkType::InfoString:
			file.Skip(file.ReadUint32BE());
			break;

		// Unrecognized chunk/value type
		default:
			ok = false;
			break;
		}
	}

	if(!m_nChannels)
		return false;

	// Channel panning
	InitializeChannels();
	for(CHANNELINDEX chn = 0; chn < m_nChannels; chn++)
	{
		ChnSettings[chn].nPan = (chn & 1) ? 256 : 0;
	}

	// convert instruments
	if(ok && (loadFlags & loadSampleData))
	{
		for(INSTRUMENTINDEX ins = 0; ins < instruments.size(); ins++)
		{
			SymInstrument &symInst = instruments[ins];

			// don't convert empty instruments (all valid instruments have a name)
			if(!symInst.HasName() || symInst.type < 0)
			{
				instrumentMap.push_back(0);
				continue;
			}
			INSTRUMENTINDEX index = GetNextFreeInstrument();
			// create this instrument and assign a sample
			if(index != INSTRUMENTINDEX_INVALID && ins < sampleMap.size())
			{
				instrumentMap.push_back(index);
				m_nInstruments = std::max<uint16>(m_nInstruments, index);

				SAMPLEINDEX sampleIndex = sampleMap[ins];
				if (symInst.IsVirtual())
				{
					// "Virtual instruments" allow an arbitrary number of other instruments to be mixed
					// at various pitches/volume levels (to create layered samples, chords, etc.)
					// Theoretically, we could handle most of these by doing some basic mixing here,
					// but for now, let's just use the first base instrument on the list
					SymEvent &event = symInst.virt.events[0];
					
					// we should be able to assume that all the base instruments will come before this one
					if(INSTRUMENTINDEX nBaseInst = event.inst; nBaseInst < instrumentMap.size())
					{
						SymInstrument &baseInst = instruments[nBaseInst];
						SAMPLEINDEX baseSample = GetNextFreeSample();
						if (baseSample != SAMPLEINDEX_INVALID)
						{
							m_nSamples++;
							ReadSampleFromSong(baseSample, *this, sampleMap[nBaseInst]);
							sampleMap[ins] = sampleIndex = baseSample;
						} else
						{
							sampleIndex = sampleMap[nBaseInst];
						}

						// correct pitch
						symInst.downsample += baseInst.downsample;
					}
				}

				ModInstrument *instr = AllocateInstrument(index, sampleIndex);
				ModSample &sample = Samples[sampleIndex];

				symInst.ConvertToMPT(*instr, sample, *this);
			} else
			{
				instrumentMap.push_back(0);
			}
		}
	}

	// convert patterns
	if(ok && (loadFlags & loadPatternData) && trackLen > 0)
	{
		// map Symphonie positions to converted patterns
		std::map<SymPosition, PATTERNINDEX> patternMap;
		// map DSP commands to MIDI macro numbers
		std::map<SymEvent, uint8> macroMap;

		const uint32 patternSize = m_nChannels * trackLen;
		PATTERNINDEX numPatterns = mpt::saturate_cast<PATTERNINDEX>(patternData.size() / patternSize);

		Patterns.ResizeArray(numPatterns);
		Order().clear();

#ifdef MPT_SYMMOD_USE_REAL_SUBSONGS
		Order.SetSequence(0);
#endif // MPT_SYMMOD_USE_REAL_SUBSONGS

		for(uint32 nSeq = 0; nSeq < sequences.size(); nSeq++)
		{
			SymSequence &seq = sequences[nSeq];
			if(seq.info == 1)
				continue;
			if(seq.info == -1)
				break;
			if(seq.start >= positions.size()
			   || seq.length > positions.size()
			   || positions.size() - seq.length < seq.start)
				continue;
			auto seqPositions = mpt::as_span(positions).subspan(seq.start, seq.length);

			// separate sequences
#ifdef MPT_SYMMOD_USE_REAL_SUBSONGS
			SEQUENCEINDEX seqIndex = Order.AddSequence(false);
			if(seqIndex == SEQUENCEINDEX_INVALID)
				break;
			Order().clear();
			ModSequence &order = Order.GetSequence(static_cast<SEQUENCEINDEX>(seqIndex));
#else
			ModSequence &order = Order();
			if (order.GetLength() > 0)
				order.push_back(ModSequence::GetInvalidPatIndex());
#endif // MPT_SYMMOD_USE_REAL_SUBSONGS

			struct ChnState
			{
				float curVolSlide = 0;       // Current volume slide factor of a channel
				float curVolSlideAmt = 0;    // Cumulative volume slide amount
				float curPitchSlide = 0;     // Current pitch slide factor of a channel
				float curPitchSlideAmt = 0;  // Cumulative pitch slide amount
				bool stopped = false;        // Sample paused or not (affects volume and pitch slides)
				uint8 lastNote = 0;          // Last note played on a channel
				int8 lastInst = 0;           // Last instrument played on a channel
				uint8 lastVol = 0;           // Last specified volume of a channel (to avoid excessive Mxx commands)
				uint8 currVibrato = 0;
				uint8 currTremolo = 0;
			};
			std::vector<ChnState> chnState(m_nChannels);

			for(auto &pos : seqPositions)
			{
				// before checking the map, apply the sequence transpose value
				pos.transpose += seq.transpose;

				// pattern already converted?
				PATTERNINDEX patternIndex;
				if(patternMap.count(pos))
				{
					patternIndex = patternMap[pos];
				} else
				{
					// convert pattern now
					patternIndex = Patterns.InsertAny(pos.length);
					if(patternIndex == PATTERNINDEX_INVALID)
						break;

					patternMap[pos] = patternIndex;

					// actually convert the pattern
					if(pos.pattern >= numPatterns)
						continue;

					uint8 patternSpeed = static_cast<uint8>(pos.speed);

					SymEvent *srcEvent = patternData.data() + (pos.pattern * patternSize) + (pos.start * m_nChannels);
					for(ROWINDEX row = 0; row < pos.length; row++)
					{
						ModCommand *rowBase = Patterns[patternIndex].GetpModCommand(row, 0);

						for(CHANNELINDEX chn = 0; chn < m_nChannels; chn++)
						{
							ModCommand &m = rowBase[chn];

							int8 note = srcEvent->note + 24; // TODO: 24 or 25? (see TODO about sample tuning)
							uint8 origInst = srcEvent->inst;
							uint8 mappedInst = 0;
							if(origInst < instruments.size())
							{
								mappedInst = static_cast<uint8>(instrumentMap[origInst]);
								if(!(instruments[origInst].instFlags & SymInstrument::NoTranspose))
									note += static_cast<int8>(pos.transpose);
							}

							switch(srcEvent->command)
							{
							case SymEvent::KeyOn:
								if(srcEvent->param > SymEvent::VolCommand)
								{
									switch (srcEvent->param)
									{
									case SymEvent::StopSample:
										m.volcmd = VOLCMD_VOLUME;
										m.vol = 0;
										chnState[chn].stopped = true;
										break;

									case SymEvent::ContSample:
										m.volcmd = VOLCMD_VOLUME;
										m.vol = 64;
										chnState[chn].stopped = false;
										break;

									case SymEvent::KeyOff:
										m.volcmd = VOLCMD_OFFSET;
										m.vol = 1;
										break;

									case SymEvent::SpeedDown:
										if (patternSpeed > 1)
										{
											m.command = CMD_SPEED;
											m.param = --patternSpeed;
										}
										break;

									case SymEvent::SpeedUp:
										if (patternSpeed < 0xFF)
										{
											m.command = CMD_SPEED;
											m.param = ++patternSpeed;
										}
										break;

									case SymEvent::SetPitch:
										m.note = chnState[chn].lastNote = note;
										// some modules (like "endless ways") use this even if a prior note hasn't played yet
										// so be sure to set instrument too (but not if inst is 0?)
										if(srcEvent->inst)
											m.instr = chnState[chn].lastInst = mappedInst;
										m.command = CMD_TONEPORTAMENTO;
										m.param = 0xFF;
										chnState[chn].curPitchSlide = 0;
										break;

									// fine portamentos with range up to half a semitone
									case SymEvent::PitchUp:
										m.command = CMD_PORTAMENTOUP;
										m.param = 0xF2;
										break;
									case SymEvent::PitchDown:
										m.command = CMD_PORTAMENTODOWN;
										m.param = 0xF2;
										break;
									case SymEvent::PitchUp2:
										m.command = CMD_PORTAMENTOUP;
										m.param = 0xF4;
										break;
									case SymEvent::PitchDown2:
										m.command = CMD_PORTAMENTODOWN;
										m.param = 0xF4;
										break;
									case SymEvent::PitchUp3:
										m.command = CMD_PORTAMENTOUP;
										m.param = 0xF8;
										break;
									case SymEvent::PitchDown3:
										m.command = CMD_PORTAMENTODOWN;
										m.param = 0xF8;
										break;
									}
								} else
								{
									if(srcEvent->note >= 0 || srcEvent->param < 100)
									{
										if(srcEvent->note >= 0)
										{
											m.note = chnState[chn].lastNote = note;
											m.instr = chnState[chn].lastInst = mappedInst;
											chnState[chn].curPitchSlide = 0;
										}

										if(srcEvent->param > 0)
										{
											uint8 newVol = mpt::saturate_round<uint8>(srcEvent->param * 0.64);

											if(chnState[chn].lastVol != newVol || chnState[chn].curVolSlide != 0)
											{
												m.command = CMD_CHANNELVOLUME;
												m.param = chnState[chn].lastVol = newVol;
											}
											chnState[chn].curVolSlide = 0;
										}
									}
								}

								// key on commands with stereo instruments are played on both channels
								if (srcEvent->note > 0
									&& (instruments[origInst].channels == SymInstrument::StereoL
										|| instruments[origInst].instFlags & SymInstrument::SyncPlay)
									&& (chn < m_nChannels - 1))
								{
									ModCommand &next = rowBase[chn + 1];
									next = m;
									if(instruments[origInst].channels == SymInstrument::StereoL)
										next.instr++;

									chnState[chn + 1].lastVol = chnState[chn].lastVol;
									chnState[chn + 1].curVolSlide = chnState[chn].curVolSlide;
									chnState[chn + 1].curVolSlideAmt = chnState[chn].curVolSlideAmt;
									chnState[chn + 1].curPitchSlide = chnState[chn].curPitchSlide;
									chnState[chn + 1].curPitchSlideAmt = chnState[chn].curPitchSlideAmt;
								}

								break;

								// volume effects
								// Symphonie has very fine fractional volume slides which are applied at the output sample rate,
								// rather than per tick or per row, so instead let's simulate it based on the pattern speed
								// by keeping track of the volume and using normal volume commands
								// the math here is an approximation which works fine for most songs
							case SymEvent::VolSlideUp:
								chnState[chn].curVolSlideAmt = 0;
								chnState[chn].curVolSlide = srcEvent->param * 0.0333f;
								break;
							case SymEvent::VolSlideDown:
								chnState[chn].curVolSlideAmt = 0;
								chnState[chn].curVolSlide = srcEvent->param * -0.0333f;
								break;

							case SymEvent::AddVolume:
								m.command = m.param = 0;
								break;
							case SymEvent::Tremolo:
							{
								// both tremolo speed and depth can go much higher than OpenMPT supports,
								// but modules will probably use pretty sane, supportable values anyway
								// TODO: handle very small nonzero params
								uint8 speed = std::min<uint8>(15, srcEvent->param >> 3);
								uint8 depth = std::min<uint8>(15, srcEvent->inst >> 3);
								chnState[chn].currTremolo = (speed << 4) | depth;
							}
								break;

								// pitch effects
								// Pitch slides have a similar granularity to volume slides, and are approximated
								// the same way here based on a rough comparison against Exx/Fxx slides
							case SymEvent::PitchSlideUp:
								chnState[chn].curPitchSlideAmt = 0;
								chnState[chn].curPitchSlide = srcEvent->param * 0.0333f;
								break;
							case SymEvent::PitchSlideDown:
								chnState[chn].curPitchSlideAmt = 0;
								chnState[chn].curPitchSlide = srcEvent->param * -0.0333f;
								break;

							case SymEvent::PitchSlideTo:
								m.command = m.param = 0;
								break;
							case SymEvent::AddPitch:
								// "The range (-128...127) is about 4 half notes."
								m.command = m.param = 0;
								break;
							case SymEvent::Vibrato:
							{
								// both vibrato speed and depth can go much higher than OpenMPT supports,
								// but modules will probably use pretty sane, supportable values anyway
								// TODO: handle very small nonzero params
								uint8 speed = std::min<uint8>(15, srcEvent->param >> 3);
								uint8 depth = std::min<uint8>(15, srcEvent->inst);
								chnState[chn].currVibrato = (speed << 4) | depth;
							}
								break;
							case SymEvent::AddHalfTone:
								m.note = (chnState[chn].lastNote += srcEvent->param);
								m.command = CMD_TONEPORTAMENTO;
								m.param = 0xFF;
								break;

								// DSP effects
							case SymEvent::Filter:
#ifndef NO_PLUGINS
							case SymEvent::DSPEcho:
							case SymEvent::DSPDelay:
#endif
								if(macroMap.count(*srcEvent))
								{
									m.command = CMD_MIDI;
									m.param = macroMap[*srcEvent];
								} else if(macroMap.size() < 128)
								{
									uint8 param = static_cast<uint8>(macroMap.size());

									ConvertDSP(*srcEvent, m_MidiCfg.szMidiZXXExt[param]);

									m.command = CMD_MIDI;
									m.param = macroMap[*srcEvent] = 0x80 | param;
								}
								break;

								// other effects
							case SymEvent::Retrig: // TODO: this takes 2 params apparently?
								m.command = CMD_RETRIG;
								m.param = std::min<uint8>(15, srcEvent->param);
								break;

							case SymEvent::SetSpeed:
								m.command = CMD_SPEED;
								m.param = patternSpeed = srcEvent->param;
								break;

							case SymEvent::CV:
								m.command = m.param = 0;
								break;
							case SymEvent::CVAdd:
								m.command = m.param = 0;
								break;

								// sample effects
							case SymEvent::FromAndPitch:
								chnState[chn].lastNote = note;
								m.instr = chnState[chn].lastInst = mappedInst;
								[[fallthrough]];
							case SymEvent::ReplayFrom:
								m.note = chnState[chn].lastNote;
								// don't always add the command, because often FromAndPitch is used with offset 0
								// to act as a key-on which doesn't cancel volume slides, etc
								if(srcEvent->param > 0)
								{
									m.command = CMD_OFFSETPERCENTAGE;
									m.param = srcEvent->param;
								}
								break;
							}

							srcEvent++;

							// any event which plays a note should re-enable continuous effects
							if(m.note != NOTE_NONE)
								chnState[chn].stopped = false;
							else if(chnState[chn].stopped)
								continue;

							// handle fractional volume slides
							if(chnState[chn].curVolSlide != 0)
							{
								chnState[chn].curVolSlideAmt += chnState[chn].curVolSlide * patternSpeed;
								if(!m.command)
								{
									if(patternSpeed > 1 && chnState[chn].curVolSlideAmt >= (patternSpeed - 1))
									{
										uint8 slideAmt = std::min<uint8>(15, static_cast<uint8>(chnState[chn].curVolSlideAmt / (patternSpeed - 1)));
										chnState[chn].curVolSlideAmt -= slideAmt * (patternSpeed - 1);
										// normal slide up
										m.command = CMD_CHANNELVOLSLIDE;
										m.param = slideAmt << 4;
									} else if(chnState[chn].curVolSlideAmt >= 1.0f)
									{
										uint8 slideAmt = std::min<uint8>(15, static_cast<uint8>(chnState[chn].curVolSlideAmt));
										chnState[chn].curVolSlideAmt -= slideAmt;
										// fine slide up
										m.command = CMD_CHANNELVOLSLIDE;
										m.param = (slideAmt << 4) | 0x0F;
									} else if(patternSpeed > 1 && chnState[chn].curVolSlideAmt <= -(patternSpeed - 1))
									{
										uint8 slideAmt = std::min<uint8>(15, static_cast<uint8>(-chnState[chn].curVolSlideAmt / (patternSpeed - 1)));
										chnState[chn].curVolSlideAmt += slideAmt * (patternSpeed - 1);
										// normal slide down
										m.command = CMD_CHANNELVOLSLIDE;
										m.param = slideAmt;
									} else if(chnState[chn].curVolSlideAmt <= -1.0f)
									{
										uint8 slideAmt = std::min<uint8>(14, static_cast<uint8>(-chnState[chn].curVolSlideAmt));
										chnState[chn].curVolSlideAmt += slideAmt;
										// fine slide down
										m.command = CMD_CHANNELVOLSLIDE;
										m.param = slideAmt | 0xF0;
									}
								}
							}
							// handle fractional pitch slides
							if(chnState[chn].curPitchSlide != 0)
							{
								chnState[chn].curPitchSlideAmt += chnState[chn].curPitchSlide * patternSpeed;
								if(m.command == CMD_NONE)
								{
									if(patternSpeed > 1 && chnState[chn].curPitchSlideAmt >= (patternSpeed - 1))
									{
										uint8 slideAmt = std::min<uint8>(0xDF, static_cast<uint8>(chnState[chn].curPitchSlideAmt / (patternSpeed - 1)));
										chnState[chn].curPitchSlideAmt -= slideAmt * (patternSpeed - 1);
										// normal slide up
										m.command = CMD_PORTAMENTOUP;
										m.param = slideAmt;
									} else if(chnState[chn].curPitchSlideAmt >= 1.0f)
									{
										uint8 slideAmt = std::min<uint8>(15, static_cast<uint8>(chnState[chn].curPitchSlideAmt));
										chnState[chn].curPitchSlideAmt -= slideAmt;
										// fine slide up
										m.command = CMD_PORTAMENTOUP;
										m.param = slideAmt | 0xF0;
									} else if(patternSpeed > 1 && chnState[chn].curPitchSlideAmt <= -(patternSpeed - 1))
									{
										uint8 slideAmt = std::min<uint8>(0xDF, static_cast<uint8>(-chnState[chn].curPitchSlideAmt / (patternSpeed - 1)));
										chnState[chn].curPitchSlideAmt += slideAmt * (patternSpeed - 1);
										// normal slide down
										m.command = CMD_PORTAMENTODOWN;
										m.param = slideAmt;
									} else if(chnState[chn].curPitchSlideAmt <= -1.0f)
									{
										uint8 slideAmt = std::min<uint8>(14, static_cast<uint8>(-chnState[chn].curPitchSlideAmt));
										chnState[chn].curPitchSlideAmt += slideAmt;
										// fine slide down
										m.command = CMD_PORTAMENTODOWN;
										m.param = slideAmt | 0xF0;
									}
								}
								// TODO: use volume column if effect column is occupied
								else if(m.volcmd == VOLCMD_NONE)
								{
									if(patternSpeed > 1 && chnState[chn].curPitchSlideAmt / 4 >= (patternSpeed - 1))
									{
										uint8 slideAmt = std::min<uint8>(9, static_cast<uint8>(chnState[chn].curPitchSlideAmt / (patternSpeed - 1)) / 4);
										chnState[chn].curPitchSlideAmt -= slideAmt * (patternSpeed - 1) * 4;
										m.volcmd = VOLCMD_PORTAUP;
										m.vol = slideAmt;
									} else if(patternSpeed > 1 && chnState[chn].curPitchSlideAmt / 4 <= -(patternSpeed - 1))
									{
										uint8 slideAmt = std::min<uint8>(9, static_cast<uint8>(-chnState[chn].curPitchSlideAmt / (patternSpeed - 1)) / 4);
										chnState[chn].curPitchSlideAmt += slideAmt * (patternSpeed - 1) * 4;
										m.volcmd = VOLCMD_PORTADOWN;
										m.vol = slideAmt;
									}
								}
							}
							// vibrato and tremolo
							if(m.command == CMD_NONE && chnState[chn].currVibrato != 0)
							{
								m.command = CMD_VIBRATO;
								m.param = chnState[chn].currVibrato;
							}
							if(m.command == CMD_NONE && chnState[chn].currTremolo != 0)
							{
								m.command = CMD_TREMOLO;
								m.param = chnState[chn].currTremolo;
							}
						}
					}

					// write speed and loop commands
					Patterns[patternIndex].WriteEffect(EffectWriter(CMD_SPEED, static_cast<uint8>(pos.speed)).Row(0).RetryNextRow());

					if(pos.loopNum > 1)
					{
						Patterns[patternIndex].WriteEffect(EffectWriter(CMD_S3MCMDEX, static_cast<ModCommand::PARAM>(0xB0 | std::min(15, pos.loopNum - 1)))
							.Row(static_cast<ROWINDEX>(pos.length - 1)));
					}
				}
				/*
				for (int i = 0; i < pos.loopNum; i++)
					*/
					order.push_back(patternIndex);
				// undo transpose tweak
				pos.transpose -= seq.transpose;
			}
		}
	}

	m_modFormat.formatName = U_("Symphonie");
	m_modFormat.type = U_("symmod");
	m_modFormat.madeWithTracker = U_("Symphonie Pro");
	m_modFormat.charset = mpt::Charset::ISO8859_1;

	return ok;
}

OPENMPT_NAMESPACE_END
